#include <unistd.h>
#include <sys/stat.h>
#include <signal.h>
#include <ktopwidget.h>
#include <ktablistbox.h>
#include <kapp.h>
#include <kmsgbox.h>
#include <kwm.h>
#include <kfiledialog.h>
#include <qwidget.h>
#include <qmlined.h>
#include <kkeyconf.h>
#include "ktail.h"
#include "getcmdline.h"

#include "ktail.moc"

#define M_HSTATUS	1
#define M_HMENU		2
#define M_HELP		3

#define REGFILE		1
#define SPECIALFILE	2
#define IPCPIPE		3

#define max(x,y)	((x)>(y))?(x):(y)

#define ALLOW_DECOR_CHANGE 1

Window mywindow;

ktail :: ktail(KApplication *, const char *name)
	: KTopLevelWidget(name)
{
char *p;

	QString tmp;
	int ii,found;

	filel.clear();

	dnd=0; tabctl=0; reading=FALSE; do_stop=FALSE;
	cmdlinedlg=0;

	menubar=new KMenuBar(this);
	fileMenu=new QPopupMenu;
	menubar->insertItem(i18n("&File"),fileMenu);
	fileMenu->insertItem(i18n("&Add file ..."),
		this,SLOT(addfiledlg()));
	fileMenu->insertItem(i18n("&Add pipe ..."),
		this,SLOT(addcmdline()));
	fileMenu->insertItem(i18n("&Remove file ..."),
		this,SLOT(removefile()));
	fileMenu->insertSeparator();
	fileMenu->insertItem(i18n("&Quit"),
		this,SLOT(close()),CTRL + Key_Q);
	optionMenu=new QPopupMenu;
	menubar->insertItem(i18n("&Options"),optionMenu);
	optionMenu->insertItem(i18n("&Font ..."),
		this,SLOT(selectfont()));
	optionMenu->insertItem(i18n("&Hide Menu"),M_HMENU);
	optionMenu->insertItem(i18n("&Hide Statusbar"),M_HSTATUS);
	
	connect(optionMenu,SIGNAL(activated(int)),this,
		SLOT(menucallback(int)));

	QString at="ktail 0.3.6";
	at+=i18n("\n\nRolf Jakob (rjakob@duffy1.franken.de)\n\nfile monitor for the KDE Desktop Environment");
	helpMenu=kapp->getHelpMenu(true,at.data());
	menubar->insertItem(i18n("&Help"),helpMenu);
	menubar->show();
	setMenu(menubar);

	KConfig *config=KApplication::getKApplication()->getConfig();
	config->setGroup("KTail");

	if (!config->hasKey("font")) {
		QFont fo;
		config->writeEntry("font",fo);
		config->sync();
		}

	ii=0; found=0;

	statusbar=new KStatusBar(this);
	statusbar->insertItem("000",STA_LINES);
	statusbar->insertItem(i18n("initialize"),STA_NAME);
	statusbar->show();
	setStatusBar(statusbar);

	active_tab=0;
	w=new mybox(this);

	createtextbox();

	QStrList files;
	if (config->hasKey("files")) {
		config->readListEntry("files",files);
		p=files.first();
		while(p) {
			addfile(p);
			p=files.next();
			}
		}

	myfont=config->readFontEntry("font");
	setfont(myfont);
	tabctl->show();

	w->adjustSize();
	w->show();

	readintextlist();
	active_tab=0;

	tim=new QTimer(this);
	connect(tim,SIGNAL(timeout()),this,SLOT(readintextlist()));
	tim->start(2000);

	setView(w);

	if (config->hasKey("statusbar"))
		if (config->readNumEntry("statusbar")==0)
			statusbar->hide();

	if (config->hasKey("menubar"))
		if (config->readNumEntry("menubar")==0)
			menubar->hide();

	if (config->hasKey("geometry")) {
		resize(config->readSizeEntry("geometry"));
		}
	prev_width=width();
	prev_height=height();

}

void ktail :: createtextbox()
{
int i;
file_str *p;

	if (!tabctl) tabctl=new KTabCtl(w);
	w->setChild(tabctl);
	w->resize(w->sizeHint());
	w->show();

	if (filel.count()) {
		p=filel.first();
		i=0;
		while(p) {
			p->textlist=new mytext(w);
			p->textlist->show();
			p->textlist->setReadOnly(TRUE);
			connect(p->textlist,SIGNAL(rightclicked()),
				this,SLOT(togglemenubar()));
			connect(p->textlist,SIGNAL(altrightclicked()),
				this,SLOT(toggledecoration()));
			connect(p->textlist,SIGNAL(middleclicked()),
				this,SLOT(togglestatusbar()));
			tabctl->addTab(p->textlist,p->filename);
			p=filel.next();
			i++;
			}
	}

	connect(tabctl,SIGNAL(tabSelected(int)),
		this,SLOT(tabchg(int)));

	dnd=new KDNDDropZone(tabctl,DndURL);
	connect(dnd,SIGNAL(dropAction(KDNDDropZone *)),
	this,SLOT(dropslot(KDNDDropZone *)));
}

void ktail :: close()
{
	closeEvent((QCloseEvent *)0);
}

void ktail :: closeEvent(QCloseEvent *e)
{
	saveconfig();
	(void)e;
	delete this;
	kapp->quit();
}

void ktail :: saveconfig()
{
file_str *help;
char line[1024];

	KConfig *config=KApplication::getKApplication()->getConfig();
	config->setGroup("KTail");

	QStrList files;
	help=filel.first();
	while(help) {
		if (help->type==IPCPIPE) {
			sprintf(line,"|%s",help->filename);
			files.append(line);
			}
		else files.append(help->filename);
		help=filel.next();
		}
	config->writeEntry("files",files);
	config->writeEntry("geometry",size());
	config->sync();
}

ktail :: ~ktail()
{
file_str *help;

	stopall();

	help=filel.first();
	while(help) {
		if (help->type==IPCPIPE && help->pid) {
			printf("killing process %d\n",help->pid);
			::kill(help->pid,SIGINT);
			}
		if (help->filep>=0) {
			::close(help->filep);
			}
		help=filel.next();
		}
}

void ktail :: stopall()
{
	tim->stop();
	if (reading) do_stop=TRUE;
}

void ktail :: togglestatusbar()
{
	KConfig *config=KApplication::getKApplication()->getConfig();
	config->setGroup("KTail");
	if (statusbar->isVisible()) {
		statusbar->hide();
		optionMenu->changeItem(i18n("&Show Statusbar"),
			M_HSTATUS);
		config->writeEntry("statusbar",0);
		} else {
		statusbar->show();
		optionMenu->changeItem(i18n("&Hide Statusbar"),
			M_HSTATUS);
		config->writeEntry("statusbar",1);
		}
	emit resizeEvent((QResizeEvent *)0);
	config->sync();
}

void ktail :: togglemenubar()
{
	KConfig *config=KApplication::getKApplication()->getConfig();
	config->setGroup("KTail");
	if (menubar->isVisible()) {
		menubar->hide();
		optionMenu->changeItem(i18n("&Show Menubar"),
			M_HMENU);
		if (!config->hasKey("menubar"))
			KMsgBox::message(this,i18n("ktail Usage"),
			i18n("You may \
reactivate the menu \n\
bar by a right click into the text window"));
		config->writeEntry("menubar",0);
		} else {
		menubar->show();
		optionMenu->changeItem(i18n("&Hide Menubar"),
			M_HMENU);
		config->writeEntry("menubar",1);
		}
	emit resizeEvent((QResizeEvent *)0);
	config->sync();
}

void ktail :: toggledecoration()
{
	long wnow=KWM::getDecoration(mywindow);
	if (wnow==KWM::normalDecoration) {
		prev_width=width();
		prev_height=height();
		KWM::setDecoration(mywindow,KWM::noDecoration);
	} else	{
		KWM::setDecoration(mywindow,KWM::normalDecoration);
		resize(prev_width,prev_height);
		}
	show();
}

void ktail :: menucallback(int no)
{
	switch(no) {
	case M_HSTATUS:
		togglestatusbar(); break;
	case M_HMENU:
		togglemenubar(); break;
	}
}

void ktail :: dropslot(KDNDDropZone *z)
{
char filen[1024];
int typ,rc;

	typ=z->getDataType();
	if (typ==DndURL) {
		strcpy(filen,z->getData());
		if (!strncmp(filen,"file:",5)) {
			rc=addfile(filen+5);
			}
		}
}


void ktail :: tabchg(int i)
{
char line[1024];
file_str *help;

	if ((unsigned)i<filel.count() && i>=0) {
		active_tab=i;
		help=filel.at(i);

		switch (help->type) {
		case SPECIALFILE:
			sprintf(line,i18n("monitoring special file %s"),
			filel.at(i)->filename);
			break;
		case IPCPIPE:
			sprintf(line,i18n("monitoring command %s"),
			filel.at(i)->filename);
			break;
		case REGFILE:
		default:
			sprintf(line,i18n("monitoring file %s"),
			help->filename);
			break;
		}

 		statusbar->changeItem(line,STA_NAME);
		sprintf(line,"%3d",help->textlist->numLines());
		statusbar->changeItem(line,STA_LINES);
		}
}

void ktail :: selectfont()
{
	KConfig *config=kapp->getConfig();
	config->setGroup("KTail");
	KFontDialog::getFont(myfont);
	config->writeEntry("font",myfont);
	config->sync();
	setfont(myfont);
}

void ktail :: addcmdline()
{
	if (!cmdlinedlg) {

	cmdlinedlg=new kgetcmdlinedlg(0,
		i18n("Enter command line to open with a pipe"),FALSE);
	cmdlinedlg->setlabel(i18n("Command :"));
	cmdlinedlg->show();
	connect(cmdlinedlg,SIGNAL(dialog_closed(bool)),
		this,SLOT(gotcmdline(bool)));
	}
}

void ktail :: gotcmdline(bool flag)
{
char line[1024],line2[1024];

	printf("got dialog_closed signal\n");
	if (cmdlinedlg && flag) {
		printf("flag is ok\n");
		strcpy(line,cmdlinedlg->getvalue());
		printf("value is %s\n",line);
		if (*line!='|') {
			strcpy(line2+1,line);
			*line2='|';
			strcpy(line,line2);
			}
		printf("adding : \"%s\"\n",line);
		addfile(line);
		}
	if (cmdlinedlg) delete cmdlinedlg;
	cmdlinedlg=0;
}

void ktail :: addfiledlg()
{
int rc;
char path[1024],*p;

	KConfig *config=kapp->getConfig();
	config->setGroup("KTail");

	if (config->hasKey("lastuseddir"))
		strcpy(path,config->readEntry("lastuseddir"));
	else	strcpy(path,"");

	QString addfilename(KFileDialog::getOpenFileName(*path?path:0,
		"*",0,i18n("ktail: Open...")) );

	if (!addfilename.isEmpty()) {
	rc=addfile(addfilename.data());
	if (rc>=0) {
		config->setGroup("KTail");
		strcpy(path,addfilename.data());
		p=strrchr(path,'/'); if (p) *p=0;
		config->writeEntry("lastuseddir",path);
		config->sync();
		}
	}
}

int ktail :: addfile(char *pfilename)
{
struct stat statbuffer;
file_str *help=new file_str;
int rc;
char *p;

	if (*pfilename) {
		help->type=0;
		strcpy(help->filename,pfilename);

		if (*pfilename=='|') {
			strcpy(help->filename,pfilename+1);
			help->type=IPCPIPE;
			}

		if (!help->type) {
			rc=stat(help->filename,&statbuffer);
			if (!rc) {
				if(S_ISREG(statbuffer.st_mode))
					help->type=REGFILE;
				else if(S_ISCHR(statbuffer.st_mode) ||
					S_ISBLK(statbuffer.st_mode) ||
					S_ISFIFO(statbuffer.st_mode) ||
					S_ISSOCK(statbuffer.st_mode))
					help->type=SPECIALFILE;
				}
			/* file does not exist, assume regular file */
			else help->type=REGFILE;
			}

		p=strrchr(help->filename,'/');
		if (p) strcpy(help->caption,p+1);
		else strcpy(help->caption,help->filename);

		help->filep=-1;
		help->fstatus=0;
		help->filepos=0;
		help->pid=0;
		help->textlist=new mytext(w);
		help->textlist->show();
		help->textlist->setReadOnly(TRUE);
		connect(help->textlist,SIGNAL(rightclicked()),
			this,SLOT(togglemenubar()));
		connect(help->textlist,SIGNAL(altrightclicked()),
			this,SLOT(toggledecoration()));
		connect(help->textlist,SIGNAL(middleclicked()),
			this,SLOT(togglestatusbar()));
		tabctl->addTab(help->textlist,help->caption);

		filel.append(help);
		setfont(myfont);
		return(1);
		}
	return(-1);
}

void ktail :: removeit(int no,int)
{
int i;
file_str *help;

	QString number=kt->text(no,0);
	QString file=kt->text(no,1);
		
	i=atoi(number.data())-1;

	if (i<0 || (unsigned)i>=filel.count()) {
		printf("trying to remove non-existant filel(%d)\n",i);
		delete kt;
		return;
		}

	help=filel.at(i);
	if (!help) {
		printf("filel.at(%d) is NULL !!\n",i);
		delete kt;
		return;
		}

	stopall();

	/* be sure that the file to be deleted is not read */
	while (reading) {
		do_stop=TRUE;
		kapp->processEvents();
		usleep(500000);
		}

	delete filel.at(i)->textlist;
	help=filel.at(i);
	if (help->type==IPCPIPE && help->pid) {
		printf("killing process %d\n",help->pid);
		::kill(help->pid,SIGINT);
		}
	filel.remove(i);
	delete help;

	help=filel.first();
	while(help) {
		if (help->textlist) {
			delete help->textlist;
			help->textlist=0;
			}
		if (help->type==IPCPIPE && help->pid) {
			printf("killing process %d\n",help->pid);
			::kill(help->pid,SIGINT);
			}
		if (help->filep>=0) {
			::close(help->filep);
			}
		help->filep=-1;
		help=filel.next();
		}

	delete dnd; dnd=0;
	w->removeChild();
	delete tabctl; tabctl=0;
	createtextbox();
	setfont(myfont);
	tabctl->show();
	do_stop=FALSE;
	tim->start(2000);

	delete kt;
	show();
}

void ktail :: removefile()
{
int i;
char tmp[1024];
file_str *p;

	kt=new KTabListBox(0,"",2);
	kt->setSeparator(' ');
	kt->setColumn(0,i18n("No."),20);
	kt->setColumn(1,i18n("File"),100);
	kt->setAutoUpdate(TRUE);
	p=filel.first(); i=1;
	while(p) {
		sprintf(tmp,"%d %s",i,p->filename);
		kt->insertItem(tmp);
		p=filel.next();
		i++;
		}
	setallcwidth(kt);
	kt->show();
	connect(kt,SIGNAL(highlighted(int,int)),this,SLOT(removeit(int,int)));
}

void ktail :: setallcwidth(KTabListBox *ktab)
{
int i;

	for (i=0;i<ktab->numCols();i++)
		setcwidth(ktab,i);
}

void ktail :: setcwidth(KTabListBox *ktab, int col)
{
int i,wi;
char tmp[1024];

	wi=0;
	QFontMetrics mymetric=kapp->generalFont;

	for (i=0;i<ktab->numRows();i++) {
		strcpy(tmp,ktab->text(i,col).data());
		wi=max(wi,mymetric.width(tmp)+15);
		}
	ktab->setColumnWidth(col,wi);
}

void ktail :: setfont(QFont fo)
{
unsigned int i;
	for (i=0;i<filel.count();i++)
		if (filel.at(i)->textlist) {
			filel.at(i)->textlist->setFont(fo);
			}
}

void ktail :: readintextlist()
{
file_str *i;

	if (reading) return;
	reading=TRUE;
	i=filel.first();
	while(i) {
		readintextlistno(i);
		if (do_stop) break;
		i=filel.next();
		}
	reading=FALSE;
	do_stop=FALSE;
}

void ktail :: openregfile(file_str *f)
{
char *line=new char[2050];
long fpos;

	if (f->filename) {
		sprintf(line,i18n("trying to open %s"),
			f->filename);
		statusbar->changeItem(line,STA_NAME);
		f->filep=open(f->filename,O_RDONLY);
		if (f->filep<0) {
			if (f->fstatus>=0) {
			f->fstatus=-1;
			sprintf(line,i18n("Could not open regular file %s:\n%s"),
				f->filename,strerror(errno));
			KMsgBox::message(this,
				i18n("ktail Error"),line);
				}

			sprintf(line,i18n("%s not opened"),
				f->filename);
			statusbar->changeItem(line,STA_NAME);
			} else {
			fpos=lseek(f->filep,0,SEEK_END);
			if (fpos<2048) fpos=0;
			else	fpos-=2048;
			fpos=lseek(f->filep,fpos,SEEK_SET);
			f->filepos=fpos;
			if (f->fstatus<0) {
			sprintf(line,i18n("File %s now open"),
				f->filename);
			f->fstatus=0;
			KMsgBox::message(this,
				i18n("ktail Info"),line);
				}
			}
		}
}

void ktail :: openspecialfile(file_str *f)
{
char *line=new char[2050];

	if (f->filename) {
		sprintf(line,i18n("trying to open %s"),
			f->filename);
		statusbar->changeItem(line,STA_NAME);
		f->filep=open(f->filename,O_RDONLY);
		if (f->filep<0) {
			if (f->fstatus>=0) {
			f->fstatus=-1;
			sprintf(line,i18n("Could not open special file %s:\n%s"),
				f->filename,strerror(errno));
			KMsgBox::message(this,
				i18n("ktail Error"),line);
				}

			sprintf(line,i18n("%s not opened"),
				f->filename);
			statusbar->changeItem(line,STA_NAME);
			} else {
			if (f->fstatus<0) {
			sprintf(line,i18n("File %s now open"),
				f->filename);
			f->fstatus=0;
			KMsgBox::message(this,
				i18n("ktail Info"),line);
				}
			}
		}
}

void ktail :: openipcpipe(file_str *f)
{
pid_t pid;
int p[2],rc;
char cmdline[1024];

	rc=::pipe(p);
	if (rc) return;
	printf("pipe successful\n");
	rc=::fcntl(p[0],F_SETFL,O_NONBLOCK);
	if (rc) perror("fcntl");
	printf("will open sh -c \"%s\"\n",f->filename);
	pid=::fork();
	if (pid<0) return;
	printf("fork successful\n");
	if (!pid) {
		::close(0);
		::open("/dev/null",O_RDONLY);
		::close(p[0]);
		::close(1);
		::dup(p[1]);
		::close(2);
		::dup(1);
//		fprintf(stdout,"stdout test\n");
//		fprintf(stderr,"stderr test\n");
		sprintf(cmdline,"%s",f->filename);
		::execl("/bin/sh","/bin/sh","-c",cmdline,NULL);
//		::system(f->filename);
//		printf("command exited\n");
		exit(0);

		} else {

		::close(p[1]);
		f->pid=pid;
		printf("child process is %d\n",pid);
		f->filep=p[0];
		}
}

int ktail :: readregfile(file_str *f)
{
int done,rc;
long fpos;
char *line=new char[2050];
char *p;
struct stat statbuffer;

	done=0;
	while(1) {

	if (do_stop) break;
	rc=read(f->filep,line,2048);
	if (rc<=0) {
		if (rc<0) perror(f->filename);
		(f->fstatus)++;
		if (f->fstatus>=MAX_TRIES) {
			rc=stat(f->filename,&statbuffer);
			if (rc<0) {
				::close(f->filep);
				f->filep=-1;
				f->textlist->clear();
				rc=0; break;
				}
			fpos=lseek(f->filep,0,SEEK_END);

			f->fstatus=0;
			if (fpos<f->filepos) {
				::close(f->filep);
				f->textlist->clear();
				f->filep=-1;
				rc=0; break;
				}
			}
			break;
		}
		f->fstatus=0;
		*(line+rc)=0;
		(f->filepos)+=rc;
		p=line+strlen(line)-1;
		while(*p=='\n' || *p=='\r') *p--=0;
		f->textlist->insertLine(line);
		done=1;
		kapp->processEvents();
	}

	delete line;

	return(done);
}

int ktail :: readspecialfile(file_str *f)
{
int done,rc;
char *line=new char[2050];
char *p;
struct stat statbuffer;

	done=0;
	while(1) {

	if (do_stop) break;
	rc=read(f->filep,line,2048);
	if (rc<=0) {
		if (rc<0) perror(f->filename);
		(f->fstatus)++;
		if (f->fstatus>=MAX_TRIES) {
			rc=stat(f->filename,&statbuffer);
			if (rc<0) {
				::close(f->filep);
				f->filep=-1;
				f->textlist->clear();
				rc=0; break;
				}
			f->fstatus=0;
			}
			break;
		}
		f->fstatus=0;
		*(line+rc)=0;
		(f->filepos)+=rc;
		p=line+strlen(line)-1;
		while(*p=='\n' || *p=='\r') *p--=0;
		f->textlist->insertLine(line);
		done=1;
		kapp->processEvents();
	}

	delete line;

	return(done);
}

int ktail :: readipcpipe(file_str *f)
{
int done,rc;
char *line=new char[2050];
char *p;

	done=0;
	while(1) {

	if (do_stop) break;
	rc=read(f->filep,line,2048);
	if (rc<=0) {
//		if (rc<0) perror(f->filename);
		(f->fstatus)++;
		break;
		}
	f->fstatus=0;
	*(line+rc)=0;
	(f->filepos)+=rc;
	p=line+strlen(line)-1;
	while(*p=='\n' || *p=='\r') *p--=0;
	f->textlist->insertLine(line);
	done=1;
	kapp->processEvents();
	}

	delete line;

	return(done);
}

void ktail :: readintextlistno(file_str *f)
{
int done=0;

	if (f->filep<0) {
		switch(f->type) {
			case REGFILE:
			openregfile(f); break;
			case SPECIALFILE:
			openspecialfile(f); break;
			case IPCPIPE:
			openipcpipe(f); break;
			}
		}

	if (f->filep>=0) {
		switch(f->type) {
			case REGFILE:
			done=readregfile(f); break;
			case SPECIALFILE:
			done=readspecialfile(f); break;
			case IPCPIPE:
			done=readipcpipe(f); break;
			}
		}

	if (done) {
		while (f->textlist->numLines()>MAX_LINES) {
			f->textlist->removeLine(0);
			kapp->processEvents();
			}
		f->textlist->setCursorPosition(f->textlist->numLines(),0);
		tabchg(filel.find(f));
		tabctl->show();
		}

}

void mybox :: setChild(QWidget *child)
{
	mychild=child;
}

void mybox :: removeChild()
{
	mychild=0;
}

void mybox :: resizeEvent( QResizeEvent *e)
{
	QWidget::resizeEvent(e);

	if (mychild) {
		mychild->resize(width(),height());
		mychild->show();
		}
}

void mytext :: mousePressEvent(QMouseEvent *e)
{
#ifdef ALLOW_DECOR_CHANGE
	if (e->button()==RightButton && e->state()==ControlButton)
		emit altrightclicked();
	else
#endif
	if (e->button()==RightButton)
		emit rightclicked();
	else if (e->button()==MidButton)
		emit middleclicked();
	else QMultiLineEdit::mousePressEvent(e);
}

int main(int argc, char* argv[])
{
int rc;

	KApplication *app=new KApplication(argc,argv);
	ktail *myktail=new ktail(kapp->getKApplication(),0);

	if (app->isRestored()) { // Session management rules
		if (KTopLevelWidget::canBeRestored(1))
			myktail->restore(1);
	}

	app->setMainWidget(myktail);
	myktail->show();

	mywindow=KWM::activeWindow();

	rc=app->exec();

	return(rc);
}

